--[[
******************************************************************************************
AvgItemLvl 1.5.0
Description: Displays the equipped average item level on the tooltip and character sheet.

Author: Sindragosa
******************************************************************************************
]]--

if not Details then return end

local Details = Details
local floor = floor
local format = format
local strmatch = strmatch
local tonumber = tonumber

local ItemLevel = {
	name = "Average Item Level",
	icon = "INTERFACE\\ICONS\\Spell_Holy_ChampionsGrace",
	attribute = false,
	spellid = false,
	author = "AvgItemLvl",
	desc = "",
	source = false,
	target = false,
	script_version = 1,
	percent_script = [[
		local value, top, total, combat, instance, player = ...
		local p = combat(DETAILS_ATTRIBUTE_MISC, player:name())
		local dps = p.dpstext or 0
		local dpstype = p.dpstype or ""
		return format("%s %.1f, %.1f", dpstype, dps, value/total*100)
	]],
	tooltip = [[
		--init:
		local player, combat, instance = ...

		--get the cooltip object (we dont use the convencional GameTooltip here)
		local GC = GameCooltip
		GC:SetOption ("YSpacingMod", 0)

		--Background RBG and Alpha:
		local r, g, b, a = 0, 0, 0, 0.75
		local r, g, b, a = 0.1960, 0.1960, 0.1960, 0.8697

		local p = combat(DETAILS_ATTRIBUTE_DAMAGE, player:name())

		local guid
		local dps = player.dps
		local hps = player.hps
		local dpstype = hps > dps and "HPS:" or "DPS:"
		local dmgtype = hps > dps and "Healing Done:" or "Damage Done:"
		local iLvl = player.itemlevel or 0

		GC:AddLine("Item Level:", iLvl)
		_detalhes:AddTooltipBackgroundStatusbar()
		GC:AddStatusBar(100, 1, r, g, b, a)

		GC:AddLine(dpstype, format("%.1f", player.dpstext or 0))
		_detalhes:AddTooltipBackgroundStatusbar()
		GC:AddStatusBar(100, 1, r, g, b, a)

		GC:AddLine(dmgtype, floor(p.total))
		_detalhes:AddTooltipBackgroundStatusbar()
		GC:AddStatusBar(100, 1, r, g, b, a)
	]],
	script = [[
		if not AvgItemLvlCache and not TTT_Cache then return 0, 0, 0 end

		--init:
		local combat, instance_container, instance = ...
		local total, top, amount = 0, 0, 0

		--get the misc actor container
		local misc_container = combat:GetActorList(DETAILS_ATTRIBUTE_MISC)

		--do the loop:
		for _, player in ipairs(misc_container) do
			--only player in group
			if player:IsGroupPlayer() then
				local p, guid
				local iLvl = 0
				local dps, hps = 0, 0

				local guid = strsub(player:guid() or "", 8)
				if guid ~= "" then
					p = combat(DETAILS_ATTRIBUTE_DAMAGE, player:name())
					if p then
						dps = p.last_dps or 0
					end
					p = combat(DETAILS_ATTRIBUTE_HEAL, player:name())
					if p then
						hps = p.last_hps or 0
					end

					if AvgItemLvlCache[guid] then
						iLvl = tonumber(AvgItemLvlCache[guid].itemLevel) or 0
					elseif TTT_Cache[guid] then
						iLvl = strmatch(TTT_Cache[guid], "^.-%-.-:.-:(.-):")
						iLvl = tonumber(iLvl) or 0
					end

					if iLvl > 0 then
						if top < iLvl then
							top = iLvl
						end

						player.dps = dps
						player.hps = hps
						if hps > dps then
							player.dpstext = format("%.1f", hps)
							player.dpstype = "HPS:"
						else
							player.dpstext = format("%.1f", dps)
							player.dpstype = "DPS:"
						end
						total = total + iLvl
						player.itemlevel = iLvl

						instance_container:SetValue(player, iLvl)
						amount = amount + 1
					end
				end
			end
		end

		--return:
		return total, top, amount
	]],
	total_script = [[
		local value, top, total, combat, instance = ...
		return format("%.2f", value)
	]],
}

Details:InstallCustomObject(ItemLevel)
