--[[
******************************************************************************************
AvgItemLvl 1.5.0
Description: Displays the equipped average item level on the tooltip and character sheet.

Author: Sindragosa
******************************************************************************************
]]--

if not Recount then return end

local Recount = Recount
local floor = floor
local format = format
local strmatch = strmatch
local tonumber = tonumber

local AvgItemLvl_Recount = CreateFrame("Frame")

local realm = GetRealmName()

------------------------------Recount Mode--------------------------------------
function AvgItemLvl_Recount:DataModes(data, num)
	if not data or not data.unit then return 0, 0 end
	local target = data.unit
	local guid = strsub(data.GUID or "", 8)
	if not guid then return 0, 0 end

	local damage, dps = Recount:MergedPetDamageDPS(data, Recount.db.profile.CurDataSet)
	local hps = (data.Fights[Recount.db.profile.CurDataSet].Healing or 0) / ((data.Fights[Recount.db.profile.CurDataSet].ActiveTime or 1) + 0)
	local tps = 0
	local ttps = ""
	if hps > dps then
		tps = format("%.1f", hps)
		ttps = "HPS: "
	else
		tps = format("%.1f", dps)
		ttps = "DPS: "
	end

	if AvgItemLvlCache[guid] then
		return floor(AvgItemLvlCache[guid].itemLevel) or 0, (ttps..tps)
	else
		return 0, 0
	end
end

function AvgItemLvl_Recount:TooltipFuncs(name, data)
	if not data or not data.unit then return end
	GameTooltip:ClearLines()
	GameTooltip:AddLine(name)
	local target = data.unit
	local guid = strsub(data.GUID or "", 8)
	if not guid then return end
	local n, r = strmatch(name or "", "^(.+)%-?(.*)")
	if not r or r == "" then
		r = realm
	end
	target = n.."-"..r

	local average = "N/A"
	if AvgItemLvlCache[guid] then
		local itemLevel = tonumber(AvgItemLvlCache[guid].itemLevel)
		average = itemLevel and format("|cffffffff%.2f|r", itemLevel) or average
	end
	GameTooltip:AddLine("Average Item Level: "..average)
end

Recount:AddModeTooltip("Average Item Level", AvgItemLvl_Recount.DataModes, AvgItemLvl_Recount.TooltipFuncs, nil, nil, nil, nil)
